/* Channel-handling routines.
 *
 * Epona (c) 2000-2001 PegSoft
 * Contact us at epona@pegsoft.net
 *
 * This program is free but copyrighted software; see the file COPYING for
 * details.
 *
 * Based on the original code of Services by Andy Church.
 *
 */

#include "services.h"

#define HASH(chan)	((chan)[1] ? ((chan)[1]&31)<<5 | ((chan)[2]&31) : 0)
Channel *chanlist[1024];

Modes modedefs[] = {
#if defined(IRC_BAHAMUT) || defined(IRC_UNREAL)
	{ 'c', CMODE_c },
#endif
	{ 'i', CMODE_I },
	{ 'm', CMODE_M },
	{ 'n', CMODE_N },
	{ 'p', CMODE_P },
	{ 's', CMODE_s },
	{ 't', CMODE_T },
	{ 'r', CMODE_r },
#ifdef IRC_UNREAL
	{ 'u', CMODE_u },
	{ 'A', CMODE_A },
	{ 'C', CMODE_C },
	{ 'G', CMODE_G },
	{ 'H', CMODE_H },
	{ 'K', CMODE_K },
#endif
#if defined(IRC_UNREAL)
	{ 'O', CMODE_O },
#endif
#ifdef IRC_UNREAL
	{ 'Q', CMODE_Q },
#endif
	{ 'R', CMODE_R },
#ifdef IRC_UNREAL
	{ 'S', CMODE_S },
	{ 'V', CMODE_V },
#endif
	{ 0 }
};

ChanStatus cstatusdefs[] = {
#if defined(IRC_UNREAL)
	{ 'a', CSTATUS_PROTECT, 0, NULL },
#elif defined(IRC_ULTIMATE)
	{ 'a', CSTATUS_ADMIN, 0, &check_valid_admin },
#endif
#if defined(IRC_UNREAL)
	{ 'h', CSTATUS_HALFOP, 0, NULL },
#elif defined(IRC_ULTIMATE)
	{ 'h', CSTATUS_HALFOP, 0, &check_valid_halfop },
#endif
	{ 'o', CSTATUS_OP, CSS_BOTSERV_NO_UNSET, &check_valid_op },
#ifdef IRC_UNREAL
	{ 'q', CSTATUS_OWNER, 0, NULL },
#endif
	{ 'v', CSTATUS_VOICE, 0, NULL },
	{ 0 }
};

/*************************************************************************/

#if 0

#if defined(IRC_UNREAL)
static char *build_modes(int32 modes, char *key, int32 limit, char *flood, char *redirect)
#elif defined(IRC_ULTIMATE)
static char *build_modes(int32 modes, char *key, int32 limit, char *flood)
#else
static char *build_modes(int32 modes, char *key, int32 limit)
#endif
{
	int i;
	char start[64], s[16], mode[2];
	char *res = scalloc(BUFSIZE, 1);
	*res = 0;
	
	for (i = 0; modedefs[i].mode > 0; i++) {
		if (modes & modedefs[i].flag) {
			sprintf(mode, "%c", modedefs[i].mode);
			strcat(start, mode);
		}
	}
	
	if (limit) snprintf(s, sizeof(s), " %d", limit);
	sprintf(res, "%s%s%s%s%s%s%s%s%s%s%s%s", 
				 start,
#if defined(IRC_UNREAL) || defined(IRC_ULTIMATE)
				 (flood ? "f" : ""),
#else
				 "",
#endif
				 (key ? "k" : ""),
				 (limit ? "l" : ""),
#ifdef IRC_UNREAL
				 (redirect ? "L" : ""),
#else
				 "",
#endif
#if defined(IRC_UNREAL) || defined(IRC_ULTIMATE)
				 (flood ? " " : ""),
				 (flood ? flood : ""),
#else
				 "",
				 "",
#endif
				 (key ? " " : ""),
				 (key ? key : ""),
				 (limit ? s : ""),
#ifdef IRC_UNREAL
				 (redirect ? " " : ""),
				 (redirect ? redirect : "")
#else
				 "",
				 ""
#endif
				 );
				 
	return res;
}

#endif

/*************************************************************************/

/* Return statistics.  Pointers are assumed to be valid. */

void get_channel_stats(long *nrec, long *memuse)
{
    long count = 0, mem = 0;
    Channel *chan;
    struct c_userlist *cu;
    BanData *bd;
    int i, j;

    for (i = 0; i < 1024; i++) {
	for (chan = chanlist[i]; chan; chan = chan->next) {
	    count++;
	    mem += sizeof(*chan);
	    if (chan->topic)
		mem += strlen(chan->topic)+1;
	    if (chan->key)
		mem += strlen(chan->key)+1;
#ifdef IRC_UNREAL
		if (chan->redirect)
		mem += strlen(chan->redirect)+1;
#endif
#if defined(IRC_UNREAL) || defined(IRC_ULTIMATE)
		if (chan->flood)
		mem += strlen(chan->flood)+1;
#endif
	    mem += sizeof(char *) * chan->bansize;
	    for (j = 0; j < chan->bancount; j++) {
		if (chan->bans[j])
		    mem += strlen(chan->bans[j])+1;
	    }
	    for (cu = chan->users; cu; cu = cu->next) {
		mem += sizeof(*cu);
		if (cu->ud) {
			mem += sizeof(*cu->ud);
			if (cu->ud->lastline) mem += strlen(cu->ud->lastline)+1;
		}
	    }
	    for (bd = chan->bd; bd; bd = bd->next)
	    {
	    	if (bd->mask) mem += strlen(bd->mask)+1;
	    	mem+= sizeof(*bd);	
	    }
	}
    }
    *nrec = count;
    *memuse = mem;
}

/*************************************************************************/

/* Return the Channel structure corresponding to the named channel, or NULL
 * if the channel was not found.  chan is assumed to be non-NULL and valid
 * (i.e. pointing to a channel name of 2 or more characters). */

Channel *findchan(const char *chan)
{
    Channel *c;

    if (debug >= 3)
	log("debug: findchan(%p)", chan);
    c = chanlist[HASH(chan)];
    while (c) {
	if (stricmp(c->name, chan) == 0)
	    return c;
	c = c->next;
    }
    if (debug >= 3)
	log("debug: findchan(%s) -> %p", chan, c);
    return NULL;
}

/*************************************************************************/

/* Iterate over all channels in the channel list.  Return NULL at end of
 * list.
 */

static Channel *current;
static int next_index;

Channel *firstchan(void)
{
    next_index = 0;
    while (next_index < 1024 && current == NULL)
	current = chanlist[next_index++];
    if (debug >= 3)
	log("debug: firstchan() returning %s",
			current ? current->name : "NULL (end of list)");
    return current;
}

Channel *nextchan(void)
{
    if (current)
	current = current->next;
    if (!current && next_index < 1024) {
	while (next_index < 1024 && current == NULL)
	    current = chanlist[next_index++];
    }
    if (debug >= 3)
	log("debug: nextchan() returning %s",
			current ? current->name : "NULL (end of list)");
    return current;
}

/*************************************************************************/
/*************************************************************************/

/* This creates the channel structure (was originally in
   chan_adduser, but splitted to make it more efficient to use for
   SJOINs). */
   
Channel *chan_create(const char *chan) 
{
	Channel *c;
	Channel **list;
	
	if (debug)
	    log("debug: Creating channel %s", chan);
	/* Allocate pre-cleared memory */
	c = scalloc(sizeof(Channel), 1);
	strscpy(c->name, chan, sizeof(c->name));
	list = &chanlist[HASH(c->name)];
	c->next = *list;
	if (*list)
	    (*list)->prev = c;
	*list = c;
	c->creation_time = time(NULL);
	/* Store ChannelInfo pointer in channel record */
	c->ci = cs_findchan(chan);
	if (c->ci) {
	    /* This is a registered channel, ensure it's mode locked +r */
	    c->ci->mlock_on |= CMODE_r;
	    c->ci->mlock_off &= ~CMODE_r;	/* just to be safe */

	    /* Store return pointer in ChannelInfo record */
	    c->ci->c = c;
	}
	/* Restore locked modes and saved topic */
	check_modes(chan);
	restore_topic(chan);
	
	return c;
}

/* Add/remove a user to/from a channel, creating or deleting the channel as
 * necessary.  If creating the channel, restore mode lock and topic as
 * necessary.  Also check for auto-opping and auto-voicing. */

/* chan_adduser now encapsulates chan_adduser2, that was created mostly to 
   handle SJOINs more efficiently. chan_adduser accepts non existent 
   channels as parameters, while chan_adduser2 does not. */

void chan_adduser(User *user, const char *chan)
{
    Channel *c = findchan(chan);
    if (!c) 
    c = chan_create(chan);
    chan_adduser2(user, c);
}

void chan_adduser2(User *user, Channel *c)
{
    struct c_userlist *u;
    char *chan = c->name;	/* FIXME: remove this in the near future ;) */
    
    if (!c) return;

#ifdef IRC_UNREAL
	if (check_should_founderize(user, chan)) {
	set_chan_status(user, c, CSTATUS_OWNER);
	} else if (check_should_protect(user, chan)) {
	set_chan_status(user, c, CSTATUS_PROTECT);
    }
#endif
#ifdef IRC_ULTIMATE
	if (check_should_admin(user, chan)) {
		set_chan_status(user, c, CSTATUS_ADMIN | CSTATUS_OP);
	} else
#endif
    if (check_should_op(user, chan)) {
	set_chan_status(user, c, CSTATUS_OP);
#if defined(IRC_UNREAL) || defined(IRC_ULTIMATE)
	} else if (check_should_halfop(user, chan)) {
	set_chan_status(user, c, CSTATUS_HALFOP);
#endif
    } else if (check_should_voice(user, chan)) {
	set_chan_status(user, c, CSTATUS_VOICE);
    }
    u = scalloc(sizeof(struct c_userlist), 1);
    u->next = c->users;
    if (c->users)
	c->users->prev = u;
    c->users = u;
    u->user = user;
    c->usercount++;
    
    /* Moved from do_join to here, it's faster */
    if (c->ci && c->ci->entry_message)
	    notice(whosends(c->ci), user->nick, "%s", c->ci->entry_message);
    
    if (s_BotServ && c->ci && c->ci->bi) {
    	if (usercount(c) == BSMinUsers) bot_join(c->ci);
    	if (usercount(c) >= BSMinUsers && (c->ci->botflags & BS_GREET) && user->na && user->na->nc->greet && check_access(user, c->ci, CA_GREET)) {
    		send_cmd(c->ci->bi->nick, "PRIVMSG %s :[%s] %s", c->name, user->na->nc->display, user->na->nc->greet);
    		c->ci->bi->lastmsg = time(NULL);	
    	}
    }
}

void chan_deluser(User *user, Channel *c)
{
    struct c_userlist *u;
    BanData *bd, *next;
    int i;
    
    if (c->ci)
    update_cs_lastseen(user, c->ci);

    for (u = c->users; u && u->user != user; u = u->next)
	;
    if (!u)
	return;
    if (u->ud)
    {
    	if (u->ud->lastline) 
    		free(u->ud->lastline);
    	free(u->ud);
    }
    if (u->next)
	u->next->prev = u->prev;
    if (u->prev)
	u->prev->next = u->next;
    else
	c->users = u->next;
    free(u);
    c->usercount--;
	
    if (s_BotServ && c->ci && c->ci->bi && usercount(c) == BSMinUsers - 1)
    	send_cmd(c->ci->bi->nick, "PART %s", c->name);
    if (!c->users) {
	if (debug)
	    log("debug: Deleting channel %s", c->name);
	for (bd = c->bd; bd; bd = next)	
	{
	    if (bd->mask) free(bd->mask);
	    next = bd->next;
	    free(bd);
	}
	if (c->ci)
	    c->ci->c = NULL;
	if (c->topic)
	    free(c->topic);
	if (c->key)
	    free(c->key);
#ifdef IRC_UNREAL
	if (c->redirect)
		free(c->redirect);
#endif
#if defined(IRC_UNREAL) || defined(IRC_ULTIMATE)
	if (c->flood)
		free(c->flood);
#endif
	for (i = 0; i < c->bancount; ++i) {
	    if (c->bans[i])
		free(c->bans[i]);
	    else
		log("channel: BUG freeing %s: bans[%d] is NULL!", c->name, i);
	}
	if (c->bansize)
	    free(c->bans);

	if (c->next)
	    c->next->prev = c->prev;
	if (c->prev)
	    c->prev->next = c->next;
	else
	    chanlist[HASH(c->name)] = c->next;
	free(c);
    }
}

/*************************************************************************/

/* Handle a channel MODE command. */

void do_cmode(const char *source, int ac, char **av)
{
	int i;
    Channel *chan;
    User *user;
    char *s, *nick;
    char c;
    int add = 1;		/* 1 if adding modes, 0 if deleting */
    char *modestr = av[1];

    chan = findchan(av[0]);
    if (!chan) {
	log("channel: MODE %s for nonexistent channel %s",
					merge_args(ac-1, av+1), av[0]);
	return;
    }

    /* This shouldn't trigger on +o, etc. */
    if (strchr(source, '.') && !modestr[strcspn(modestr, "bovah")]) {
	if (time(NULL) != chan->server_modetime) {
	    chan->server_modecount = 0;
	    chan->server_modetime = time(NULL);
	}
	chan->server_modecount++;
    }

    s = modestr;
    ac -= 2;
    av += 2;

    while (*s) {
    	
    switch ((c = *s++)) {

	case '+':
	    add = 1; break;

	case '-':
	    add = 0; break;

	case 'k':
	    if (--ac < 0) {
		log("channel: MODE %s %s: missing parameter for %ck",
					chan->name, modestr, add ? '+' : '-');
		break;
	    }
	    if (chan->key) {
		free(chan->key);
		chan->key = NULL;
	    }
	    if (add)
		chan->key = sstrdup(*av++);
	    break;

	case 'l':
	    if (add) {
		if (--ac < 0) {
		    log("channel: MODE %s %s: missing parameter for +l",
							chan->name, modestr);
		    break;
		}
		chan->limit = atoi(*av++);
	    } else {
		chan->limit = 0;
	    }
	    break;
	    
#if defined(IRC_UNREAL) || defined(IRC_ULTIMATE)
	case 'f':
	    if (--ac < 0) {
		log("channel: MODE %s %s: missing parameter for %cf",
					chan->name, modestr, add ? '+' : '-');
		break;
	    }
	    if (chan->flood) {
		free(chan->flood);
		chan->flood = NULL;
	    }
	    if (add)
		chan->flood = sstrdup(*av++);
	    break;
#endif
#ifdef IRC_UNREAL
	case 'L':
	    if (--ac < 0) {
		log("channel: MODE %s %s: missing parameter for %cL",
					chan->name, modestr, add ? '+' : '-');
		break;
	    }
	    if (chan->redirect) {
		free(chan->redirect);
		chan->redirect = NULL;
	    }
	    if (add)
		chan->redirect = sstrdup(*av++);
	    break;
#endif

	case 'b':
	    if (--ac < 0) {
		log("channel: MODE %s %s: missing parameter for %cb",
					chan->name, modestr, add ? '+' : '-');
		break;
	    }
	    if (add) {
	    	char *banmask = *av++;
	    	
	    	if (s_BotServ && BSSmartJoin && chan->ci && chan->ci->bi && usercount(chan) >= BSMinUsers) {
	    		char mask[BUFSIZE];
	    		BotInfo *bi = chan->ci->bi;
	    		
	    		snprintf(mask, sizeof(mask), "%s!%s@%s", bi->nick, bi->user, bi->host);
	    		if (match_wild_nocase(banmask, mask)) {
	    			send_cmd(bi->nick, "MODE %s -b %s", chan->name, banmask);
	    			break;	
	    		}
	    	}
	    	
		if (chan->bancount >= chan->bansize) {
		    chan->bansize += 8;
		    chan->bans = srealloc(chan->bans,
					sizeof(char *) * chan->bansize);
		}
		chan->bans[chan->bancount++] = sstrdup(banmask);
	    } else {
		char **s = chan->bans;
		int i = 0;
		while (i < chan->bancount && strcmp(*s, *av) != 0) {
		    i++;
		    s++;
		}
		if (i < chan->bancount) {
		    chan->bancount--;
		    if (i < chan->bancount)
			memmove(s, s+1, sizeof(char *) * (chan->bancount-i));
		}
		av++;
	    }
	    break;
	
	default:
		for (i = 0; cstatusdefs[i].mode > 0; i++) {
			if (c == cstatusdefs[i].mode) {
				BotInfo *bi;
				
				if (--ac < 0) {
					log("channel: MODE %s %s: missing parameter for %c%c", chan->name, modestr, add ? '+' : '-', c);
					break;
	    		}
	    		nick = *av++;
	    		
	    		bi = findbot(nick);
	    		if (bi) {
	    			if ((cstatusdefs[i].special & CSS_BOTSERV_NO_UNSET)	&& !add)
	    				send_cmd(bi->nick, "MODE %s +%c %s", chan->name, c, nick);
	    			/* Don't need to continue */
	    			break;
	    		}
	    		
	    		user = finduser(nick);
	    		if (!user) {
		    		log("channel: MODE %s %c%c for nonexistent user %s", chan->name, (add ? '+' : '-'), c, nick);
		    		break;
				}
				
	    		if (add) {
	    			if (cstatusdefs[i].isvalid && !cstatusdefs[i].isvalid(user, chan->name, !!strchr(source, '.')))
                    	break;
	    			set_chan_status(user, chan, cstatusdefs[i].flag);
	    		}
	    		else remove_chan_status(user, chan, cstatusdefs[i].flag);
	    		
	    		if (debug)
		    		log("debug: Setting %c%c on %s for %s", (add ? '+' : '-'), c, chan->name, nick);
	    		
	    		break;
	    	}
		}
		
		for (i = 0; modedefs[i].mode > 0; i++) {
			if (c == modedefs[i].mode) {
				if (add) chan->mode |= modedefs[i].flag;
	    		else chan->mode &= ~modedefs[i].flag;
				break;
			}
		}
		break;


	} /* switch */

    } /* while (*s) */

    /* Check modes against ChanServ mode lock */
    check_modes(chan->name);
}

/*************************************************************************/

/* Handle a TOPIC command. */

void do_topic(const char *source, int ac, char **av)
{
    Channel *c = findchan(av[0]);

    if (!c) {
	log("channel: TOPIC %s for nonexistent channel %s",
						merge_args(ac-1, av+1), av[0]);
	return;
    }
    if (check_topiclock(av[0]))
	return;
    strscpy(c->topic_setter, av[1], sizeof(c->topic_setter));
    c->topic_time = atol(av[2]);
    if (c->topic) {
	free(c->topic);
	c->topic = NULL;
    }
    if (ac > 3 && *av[3])
	c->topic = sstrdup(av[3]);
    record_topic(av[0]);
}

/*************************************************************************/

/* Determines who must send a channel-related message when symbiosis is ON 
	--lara */

char *whosends(ChannelInfo *ci) {
	if (!ci || !(ci->botflags & BS_SYMBIOSIS) || !ci->bi || !ci->c || usercount(ci->c) < BSMinUsers) return s_ChanServ;
	else return ci->bi->nick;	
}

/*************************************************************************/
